import 'dart:async';
import 'dart:collection';
import 'dart:ui';

import 'package:eyflutter_core/eyflutter_core.dart';
import 'package:eyflutter_player/beans/player_info.dart';
import 'package:eyflutter_player/enums/play_mode.dart';
import 'package:fijkplayer/fijkplayer.dart';

class AudioUtils {
  factory AudioUtils() => _getInstance();

  static AudioUtils get instance => _getInstance();
  static AudioUtils? _instance;

  static AudioUtils _getInstance() {
    return _instance ??= AudioUtils._internal();
  }

  AudioUtils._internal();

  final LinkedHashMap<String, PlayerInfo> _items = LinkedHashMap();
  final Map<String, FijkPlayer> playMap = {};
  ListQueue<PlayerInfo> queue = ListQueue<PlayerInfo>();

  bool _isPreparing = false;
  Function()? _initPlayerCall;
  final StreamController<double> _playProgressController = StreamController.broadcast();

  //播放进度更新
  Stream<double> get onPlayProgressUpdate => _playProgressController.stream;

  /// 播放索引
  int playIndex = 0;

  PlayMode _mode = PlayMode.cycle;

  /// 播放模式
  PlayMode get mode => _mode;

  /// 设置播放模式
  void setPlayMode(PlayMode mode) {
    _mode = mode;
  }

  /// 播放列表
  List<PlayerInfo> get playList => _items.values.toList();

  PlayerInfo _getPlayInfo() {
    if (_items.isEmptyMap() || (playIndex + 1) > _items.length) {
      return PlayerInfo(playerId: "", url: "");
    }
    return playList[playIndex];
  }

  /// 获取播放信息
  PlayerInfo get playInfo => _getPlayInfo();

  /// 获取播放器
  FijkPlayer? get player => playMap[playInfo.playerId];

  FijkPlayer _instancePlayer(PlayerInfo entry) {
    return FijkPlayer()..setDataSource(entry.url);
  }

  /// 追加播放列表
  void appendPlayList({required List<PlayerInfo> list}) {
    for (var item in list) {
      if (_items.containsKey(item.playerId)) {
        continue;
      }
      queue.addFirst(item);
    }
    _startLoad();
  }

  /// 追加播放信息
  void appendPlay({required PlayerInfo playerInfo}) {
    if (!_items.containsKey(playerInfo.playerId)) {
      queue.addFirst(playerInfo);
      _startLoad();
    }
  }

  void _startLoad() {
    if (!_isPreparing) {
      _prepare();
    }
  }

  _prepare() {
    if (queue.isEmpty) {
      _isPreparing = false;
      return;
    }
    _isPreparing = true;
    var removeLast = queue.removeLast();
    if (_items.containsKey(removeLast.playerId)) {
      _prepare();
      return;
    }
    var player = _instancePlayer(removeLast);
    _initPlayerListener(() {
      if (player.state == FijkState.initialized) {
        _items[removeLast.playerId] = removeLast;
        playMap[removeLast.playerId] = player;
        player.removeListener(_initPlayerCall!);
        _isPreparing = false;
        _prepare();
      } else if (player.state == FijkState.error) {
        player.removeListener(_initPlayerCall!);
        //资源无效加载下一个
        _isPreparing = false;
        _prepare();
      }
    });
    player.addListener(_initPlayerCall!);
    player.onCurrentPosUpdate.listen((event) => _playProgressListen(player.value.duration, event));
  }

  void _initPlayerListener(VoidCallback initCall) {
    _initPlayerCall = initCall;
  }

  void _playProgressListen(Duration total, Duration duration) {
    var currentTime = duration.inMilliseconds.toDouble();
    var totalTime = total.inMilliseconds.toDouble();
    var progress = currentTime / totalTime;
    _playProgressController.add(progress);
    if (progress == 1.0) {
      _switchAudio();
    }
  }

  void _switchAudio() {
    var _player = player;
    if (_mode == PlayMode.cycle) {
      _player?.seekTo(0);
      _player?.pause();
      if ((playIndex + 1) < _items.length) {
        playIndex++;
      } else {
        playIndex = 0;
      }
      _player?.start();
    } else if (_mode == PlayMode.sequence && (playIndex + 1) < _items.length) {
      _player?.seekTo(0);
      _player?.pause();
      playIndex++;
      _player?.start();
    } else if (_mode == PlayMode.singleCycle) {
      _player?.seekTo(0);
      _player?.start();
    }
  }

  /// 开始播放
  /// [index] 播放索引
  void start({int index = 0}) {
    playIndex = index;
    var _player = player;
    var total = _player?.value.duration.inMilliseconds ?? 0;
    var milliseconds = _player?.currentPos.inMilliseconds ?? 0;
    if (milliseconds == total) {
      _player?.seekTo(0);
    } else {
      var targetSec = milliseconds - 500;
      if (targetSec > 0) {
        _player?.seekTo(targetSec);
      }
    }
    _player?.start();
  }

  /// 重置播放
  void reset({int index = 0}) {
    playIndex = index;
    var _player = player;
    _player?.seekTo(0);
    _player?.pause();
  }

  /// 暂停播放
  void pause({int index = 0}) {
    playIndex = index;
    player?.pause();
  }

  /// 释放资源
  void dispose() {
    playMap.forEach((key, value) {
      value.release();
    });
    _items.clear();
  }

  /// 切换至上一个音频
  void prev() {
    if (playIndex > 0) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      playIndex--;
      _player = player;
      _player?.start();
    }
  }

  /// 切换至下一个视频
  void next() {
    if ((playIndex + 1) < _items.length) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      playIndex++;
      _player = player;
      _player?.start();
    }
  }
}
